# -*- coding: utf-8 -*-
"""
.. module:: downloader

This module provides the downloaders able to grab content from the
supported sites.

.. moduleauthor:: Asiel Díaz Benítez <asieldbenitez@gmail.com>

"""

from abc import ABC, abstractmethod
from http.cookiejar import MozillaCookieJar
import imghdr  # type: ignore
import json
import logging
import logging.handlers
import os
import re
import typing
from urllib.request import (build_opener, HTTPCookieProcessor,  # type: ignore
                            Request, URLError)
from urllib.parse import urlencode, quote_plus

from bs4 import BeautifulSoup  # type: ignore

from smd.utils import (ConsoleFilter, Chapter, get_text, select_chapters,
                       select_mangas, Manga, mkdir, random_ua)

if typing.TYPE_CHECKING:
    from typing import Any, List

    def _(msg: str) -> str:
        return msg


class Downloader(ABC):

    """Abstract class base of all manga downloaders."""

    logfile = 'smd.log'
    verbose = False

    def __init__(self, name: str, lang: str, site_url: str) -> None:
        """Should be called by child classes.

        :param name: the name of the downloader.
        :param lang: the language of the downloader (ISO 639-1 code) or *
                     if the site have manga in multiple languages.
        :param site_url: the URL of the web site of this downloader.
        """
        self.name = name
        self.lang = lang
        self.site_url = site_url
        cookie_jar = MozillaCookieJar()
        # cookie_jar.load('cookies.txt') # TODO: save and load cookies
        self.url_opener = build_opener(HTTPCookieProcessor(cookie_jar))
        self.url_opener.addheaders = [
            ('Host', self.site_url),
            ('Referer', self.site_url),
            ('User-Agent', random_ua()),
            ('Accept', 'application/json, text/plain, */*'),
            ('Accept-Language', 'en-US,en;q=0.5'),
            ('DNT', '1'),  # Do Not Track
        ]
        self._init_logger()

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return '({}, {}, {})'.format(self.name, self.lang, self.site_url)

    def __del__(self):
        for handler in self.logger.handlers:
            handler.close()

    def _download_chapter(self, chapter: Chapter) -> None:
        """Downloads the given chapter only if it is not downloaded already,
        if the chapter was previously interrupted the download is resumed.

        :param chapter: the chapter to download.
        """
        if chapter.current == len(chapter.images):
            self.logger.info(_("Skipping chapter '{}': Already downloaded.")
                             .format(chapter))
            return
        elif chapter.current == -1:
            self.logger.info(_("Getting images list for chapter '{}' ...")
                             .format(chapter))
            chapter.images = self.get_images(chapter.url)
            chapter.current = 0
            chapter.save_data()
        img_count = len(chapter.images)
        dcount = len(str(img_count))
        for url in chapter.images[chapter.current:]:
            current = chapter.current + 1
            print('\r'+_("[{}] Downloading '{}' (image: {}/{})").format(
                self.name, chapter, current, img_count), end='')
            name = os.path.join(chapter.path, str(current).zfill(dcount))
            self.download_img(self.get_image(url), name)
            chapter.current = current
            chapter.save_data()
        if img_count > 0:
            print()

    def _init_logger(self) -> None:
        """Initializes the downloader is logger."""
        self.logger = logging.Logger(self.name)
        self.logger.parent = None  # type: ignore
        fhandler = logging.handlers.RotatingFileHandler(Downloader.logfile,
                                                        backupCount=1,
                                                        maxBytes=2000000)
        fhandler.setLevel(logging.DEBUG)
        chandler = logging.StreamHandler()
        if Downloader.verbose:
            chandler.setLevel(logging.DEBUG)
        else:
            chandler.addFilter(ConsoleFilter())
            chandler.setLevel(logging.INFO)
        formatter = logging.Formatter('%(asctime)s - %(name)s - '
                                      '%(levelname)s - %(message)s')
        fhandler.setFormatter(formatter)
        formatter = logging.Formatter('[%(name)s] - '
                                      '%(levelname)s - %(message)s')
        chandler.setFormatter(formatter)
        # order is important because ConsoleFilter:
        self.logger.addHandler(fhandler)
        self.logger.addHandler(chandler)

    def download(self, manga_name: str, chapter_selectors: str) -> bool:
        """Searches for the given manga and, if found, chapters specified in
        ``chapter_selectors`` are downloaded.

        :param manga_name: the manga to search and download.
        :param chapter_selectors: a string specifying the chapters to download.
        :return: ``True`` if the manga was downloaded successfully, ``False``
                 otherwise.
        """
        success = True
        try:
            self.logger.info(_("Searching for '{}' ...").format(manga_name))
            mangas = self.search(manga_name)
            if mangas:
                manga = select_mangas(mangas, multiple=False)[0]
            else:
                self.logger.info(_("'{}' not found.").format(manga_name))
                return False
            manga.path = mkdir(os.path.abspath('.'), manga.title)
            manga.save_data()
            self.logger.info(_("Getting chapters list of '{}' ...")
                             .format(manga))
            chapters = self.get_chapters(manga.url)
            self.logger.info(_("Found {} chapters for '{}'")
                             .format(len(chapters), manga))
            chapters = select_chapters(chapters, chapter_selectors)
            self.logger.info(_("Selected {} chapters to download")
                             .format(len(chapters)))
            self.logger.debug(_('Creating chapter folders...'))
            for chap in chapters:
                path = manga.get_new_chapter_path()
                os.mkdir(path)  # TODO: handle Exceptions
                chap.path = path
                chap.save_data()
            self.logger.info(_("Downloading '{}':").format(manga))
            for chapter in chapters:
                self._download_chapter(chapter)
        except KeyboardInterrupt:
            raise
        except URLError as ex:
            self.logger.exception(ex)
            success = False
        return success

    def download_img(self, url: str, name: str) -> str:
        """Receives an image URL and a filename (without a file extension)
        and downloads the image, detects image format and save it to the
        given filename plus the proper extension.

        :param url: the URL of the image.
        :param name: the name to use for the saved image (without file
                     extension)
        :return: the file name with file extension  of the downloaded image.
        """
        img = self.get_bytes(url)
        ext = imghdr.what("", h=img)
        ext = '' if ext is None else '.'+ext
        file_name = name+ext
        with open(file_name, 'bw') as outf:
            outf.write(img)
        return file_name

    def get_bytes(self, url: str, data: dict = None, method: str = 'GET',
            xhr: bool = False) -> bytes:
        """Retrieves data from given URL.

        :param url: the URL to retrieve.
        :param data: the data to send with the request.
        :param method: the method to use to request the URL (POST/GET).
        :param xhr: if ``True`` set the header 'X-Requested-With' to
                    'XMLHttpRequest'
        :return: the response data.
        :raises ConnectionResetError: if the connection is reset more than
                                      five times.
        :raises ValueError: if ``method`` is not ``GET`` or ``POST``.
        """
        if data is None:
            data = {}
        method = method.upper()
        encoded_data = urlencode(data)
        del data
        if xhr:
            headers = {'X-Requested-With': 'XMLHttpRequest'}
        else:
            headers = {}
        if method == 'GET':
            if encoded_data:
                url = url+'?'+encoded_data
            req = Request(url, None, headers)
        elif method == 'POST':
            req = Request(url, encoded_data.encode('ascii'), headers)
        else:
            raise ValueError(
                _("Only 'GET' and 'POST' methods are implemented."))
        errors = 0
        while True:
            try:
                self.logger.debug(_('Downloading: {}').format(url))
                with self.url_opener.open(req) as resp:
                        return resp.read()
            except ConnectionResetError as err:
                if errors > 5:
                    raise
                errors += 1
                self.logger.debug(err)

    @abstractmethod
    def get_chapters(self, manga_url: str) -> 'List[Chapter]':
        """Extracts the manga chapters from the given manga URL.

        :param manga_url: the URL of a manga.
        :return: the list of chapters of the given manga.
        """
        pass

    def get_image(self, image_url: str) -> str:
        """If needed, child classes must override this method to extracts the
        image link from the given URL.

        :param image_url: the image URL or an URL to a page where the image
                          link can be extracted.
        :return: the image URL.
        """
        return image_url

    @abstractmethod
    def get_images(self, chapter_url: str) -> 'List[str]':
        """Extracts the images links from a chapter URL.

        :param chapter_url: the URL of a manga chapter.
        :return: list of image URLs for the given chapter, or a list of
                 URLs where the images can be extracted using
                 :func:`~smd.downloader.Downloader.get_image`.
        """
        pass

    def get_json(self, *args, **kargs) -> 'Any':
        """Request json data from the given url, the given parameters are
        passed to :meth:`~smd.downloader.Downloader.get_str` method.

        :return: the response data object.
        """
        kargs['xhr'] = True
        return json.loads(self.get_str(*args, **kargs))

    def get_str(self, *args, **kargs) -> str:
        """Calls :meth:`~smd.downloader.Downloader.get_bytes` with the given
        parameters and returns the response bytes decoded.
        """
        return self.get_bytes(*args, **kargs).decode(errors='ignore')

    def resume(self, manga: Manga) -> int:
        """Continues with the (unfinished) download of the given manga.

        :param manga: the manga to continue downloading.
        :return: the number of resumed mangas.
        """
        self.logger.info(_("Resuming download of '{}' ...").format(manga))
        downloads = 0
        for chap in manga.chapters():
            if chap.current < len(chap.images):
                self._download_chapter(chap)
                downloads += 1
        if downloads == 0:
            self.logger.warning(_("No pending chapters found for '{}'.")
                                .format(manga))
        return downloads

    @abstractmethod
    def search(self, manga: str) -> 'List[Manga]':
        """Searches for mangas matching the given text.

        :param manga: a frase or manga title to search.
        :return: a list of mangas.
        """
        pass

    def update(self, manga: Manga) -> int:
        """Downloads new available chapters of the given manga.

        :param manga: the manga to update.
        :return: the number of updated mangas.
        """
        self.logger.info(_("Looking for new chapters of '{}' ...")
                         .format(manga))
        all_chaps = self.get_chapters(manga.url)
        old_chaps = [chap.url for chap in manga.chapters()]
        new_chaps = []  # type: List[Chapter]
        for chap in all_chaps:
            if chap.url not in old_chaps:
                path = manga.get_new_chapter_path()
                os.mkdir(path)
                chap.path = path
                chap.save_data()
                new_chaps.append(chap)
        self.logger.info(_("Found {} new chapters for '{}'")
                         .format(len(new_chaps), manga))
        for chap in new_chaps:
            self._download_chapter(chap)
        updates = len(new_chaps)
        if updates == 0:
            self.logger.info(_("No new chapters found for '{}'.")
                             .format(manga))
        return updates


class NineManga(Downloader):

    """Downloads manga from ninemanga.com."""

    def __init__(self, site: str) -> None:
        lang = 'pt' if site == 'br' else site
        super().__init__('ninemanga-'+site, lang,
                         'http://{}.ninemanga.com'.format(site))

    def search(self, manga: str) -> 'List[Manga]':
        url = self.site_url+"/search/"
        soup = BeautifulSoup(self.get_str(url, {'wd': manga}), 'html.parser')
        direlist = soup.find('ul', class_='direlist')
        results = [Manga('', get_text(a), a['href'], self.name)
                   for a in direlist.find_all('a', class_='bookname')]  # type: List[Manga]
        pagelist = soup.find('ul', class_='pagelist')
        if pagelist:
            # this get only first few pages:
            for page in pagelist.find_all('a')[1:-1]:
                soup = BeautifulSoup(self.get_str(page['href']), 'html.parser')
                direlist = soup.find('ul', class_='direlist')
                anchors = direlist.find_all('a', class_='bookname')
                results.extend(Manga('', get_text(a), a['href'], self.name)
                               for a in anchors)
        return results

    def get_chapters(self, manga_url: str) -> 'List[Chapter]':
        soup = BeautifulSoup(self.get_str(manga_url), 'html.parser')
        tag = soup.find('div', class_='warning')
        if tag:
            soup = BeautifulSoup(self.get_str(tag.a['href']), 'html.parser')
        tag = soup.find('div', class_='silde')
        chapters = [Chapter('', a['title'], a['href'])
                    for a in tag.find_all('a', class_='chapter_list_a')]
        chapters.reverse()
        return chapters

    def get_images(self, chapter_url: str) -> 'List[str]':
        soup = BeautifulSoup(self.get_str(chapter_url), 'html.parser')
        tag = soup.find('select', id='page')
        return [self.site_url + opt['value']
                for opt in tag.find_all('option')]

    def get_image(self, image_url: str) -> str:
        soup = BeautifulSoup(self.get_str(image_url), 'html.parser')
        return soup.find('img', class_='manga_pic')['src']


class HeavenManga(Downloader):

    """Downloads manga from heavenmanga.com."""

    def __init__(self) -> None:
        super().__init__('heavenmanga', 'es', 'http://heavenmanga.com')

    def search(self, manga: str) -> 'List[Manga]':
        # TODO: find a better way to do this:
        url = '{}/buscar/{}.html'.format(self.site_url, quote_plus(manga))
        # page restriction: len(manga) must to be >= 4
        soup = BeautifulSoup(self.get_str(url), 'html.parser')
        divs = soup.find_all('div', class_='cont_manga')
        return [Manga('', get_text(div.a.header), div.a['href'], self.name)
                for div in divs]

    def get_chapters(self, manga_url: str) -> 'List[Chapter]':
        soup = BeautifulSoup(self.get_str(manga_url), 'html.parser')
        chapters = [Chapter('', a['title'], a['href'])
                    for a in soup.find('ul', id='holder').find_all('a')]
        chapters.reverse()
        return chapters

    def get_images(self, chapter_url: str) -> 'List[str]':
        soup = BeautifulSoup(self.get_str(chapter_url), 'html.parser')
        chapter_url = soup.find('a', id='l')['href']
        soup = BeautifulSoup(self.get_str(chapter_url), 'html.parser')
        return [opt['value']
                for opt in soup.find('select').find_all('option')]

    def get_image(self, image_url: str) -> str:
        soup = BeautifulSoup(self.get_str(image_url), 'html.parser')
        return soup.find('img', id='p')['src']


class MangaReader(Downloader):

    """Downloads manga from www.mangareader.net."""

    def __init__(self) -> None:
        super().__init__('mangareader', 'en', 'https://www.mangareader.net')

    def search(self, manga: str) -> 'List[Manga]':
        url = self.site_url+'/actions/search/'
        results = []  # type: List[Manga]
        for line in self.get_str(url, {'q': manga, 'limit': 100}).splitlines():
            row = line.split('|')
            if len(row) != 6:
                self.logger.warning("unknow line format: {}"
                                    .format('|'.join(row)))
                continue
            results.append(Manga('', row[2], self.site_url+row[-2], self.name))
        return results

    def get_chapters(self, manga_url: str) -> 'List[Chapter]':
        soup = BeautifulSoup(self.get_str(manga_url), 'html.parser')
        chapters = [Chapter('', get_text(a), self.site_url + a['href'])
                    for a in soup.find('table', id='listing').find_all('a')]
        # don't need to use `chapters.reverse()` here
        return chapters

    def get_images(self, chapter_url: str) -> 'List[str]':
        soup = BeautifulSoup(self.get_str(chapter_url), 'html.parser')
        tag = soup.find('select', id='pageMenu')
        return [self.site_url + opt['value']
                for opt in tag.find_all('option')]

    def get_image(self, image_url: str) -> str:
        soup = BeautifulSoup(self.get_str(image_url), 'html.parser')
        return soup.find('img', id='img')['src']


class MangaAll(Downloader):

    """Downloads manga from mangaall.net."""

    def __init__(self) -> None:
        super().__init__('mangaall', 'en', 'http://mangaall.net')
        self.regex = re.compile(r"var _page_total = '(?P<total>[0-9]+)';")

    def search(self, manga: str) -> 'List[Manga]':
        url = self.site_url+'/search/'
        soup = BeautifulSoup(self.get_str(url, {'q': manga}), 'html.parser')
        divs = soup.find_all('div', class_='mainpage-manga')
        results = [Manga('', a['title'], a['href'], self.name)
                   for a in (div.find('div', 'media-body').a for div in divs)]
        # TODO: get other results pages
        return results

    def get_chapters(self, manga_url: str) -> 'List[Chapter]':
        soup = BeautifulSoup(self.get_str(manga_url), 'html.parser')
        chapters = [Chapter('', a['title'], a['href'])
                    for a in soup.find('section', id='examples').find_all('a')]
        chapters.reverse()
        return chapters

    def get_images(self, chapter_url: str) -> 'List[str]':
        soup = BeautifulSoup(self.get_str(chapter_url), 'html.parser')
        for script in soup.find_all('script'):
            matches = [int(n) for n in self.regex.findall(str(script.string))]
            if matches:
                return [chapter_url+'?page='+str(n)
                        for n in range(1, matches[-1]+1)]
        raise Exception("Can't find images list")

    def get_image(self, image_url: str) -> str:
        soup = BeautifulSoup(self.get_str(image_url), 'html.parser')
        return soup.find('div', class_='each-page').img['src']


class MangaDoor(Downloader):

    """Downloads manga from mangadoor.com."""

    def __init__(self) -> None:
        super().__init__('mangadoor', 'es', 'http://mangadoor.com')

    def search(self, manga: str) -> 'List[Manga]':
        url = self.site_url+'/search/'
        suggestions = self.get_json(url, {'query': manga})['suggestions']
        results = [Manga('', sugg['value'],
                         self.site_url+'/manga/'+sugg['data'], self.name)
                   for sugg in suggestions]
        return results

    def get_chapters(self, manga_url: str) -> 'List[Chapter]':
        soup = BeautifulSoup(self.get_str(manga_url), 'html.parser')
        chapters = [Chapter('', get_text(a), a['href'])
                    for a in soup.find('ul', class_='chapters').find_all('a')]
        chapters.reverse()
        return chapters

    def get_images(self, chapter_url: str) -> 'List[str]':
        soup = BeautifulSoup(self.get_str(chapter_url), 'html.parser')
        opts = soup.find('select', id='page-list').find_all('option')
        return [chapter_url+'/'+opt['value'] for opt in opts]

    def get_image(self, image_url: str) -> str:
        soup = BeautifulSoup(self.get_str(image_url), 'html.parser')
        return soup.find('div', id='ppp').img['src']


class MangaNelo(Downloader):

    """Downloads manga from manganelo.com."""

    def __init__(self) -> None:
        super().__init__('manganelo', 'en', 'https://manganelo.com')

    def search(self, manga: str) -> 'List[Manga]':
        query_str = ''
        for char in manga:
            if char.isalnum():
                query_str += char
            else:
                query_str += ' '
        query_str = '_'.join(query_str.split())
        req_data = {'search_style': 'tentruyen', 'searchword': query_str}
        url = self.site_url+'/home_json_search/'
        res_data = self.get_json(url, req_data
                                 , method='POST') # type: List[dict]
        return [Manga('', get_text(BeautifulSoup(result['name'],
                                                 'html.parser')),
                      self.site_url+'/manga/'+result['nameunsigned'], self.name)
                for result in res_data]

    def get_chapters(self, manga_url: str) -> 'List[Chapter]':
        soup = BeautifulSoup(self.get_str(manga_url), 'html.parser')
        div = soup.find('div', class_='chapter-list')
        chapters = []
        for anchor in div.find_all('a'):
            if anchor['href'].startswith('/'):
                anchor['href'] = 'https:'+anchor['href']
            chapters.append(Chapter('', get_text(anchor), anchor['href']))
        chapters.reverse()
        return chapters

    def get_images(self, chapter_url: str) -> 'List[str]':
        soup = BeautifulSoup(self.get_str(chapter_url), 'html.parser')
        div = soup.find('div', id='vungdoc')
        return [img['src'] for img in div.find_all('img')]


class MangaHere(Downloader):

    """Downloads manga from www.mangahere.cc."""

    def __init__(self) -> None:
        super().__init__('mangahere', 'en', 'http://www.mangahere.cc')

    def search(self, manga: str) -> 'List[Manga]':
        url = self.site_url+'/ajax/search.php'
        data = self.get_json(url, {'query': manga})
        results = [Manga('', title, 'http:'+url, self.name)
                   for title, url in zip(data['suggestions'], data['data'])]
        return results

    def get_chapters(self, manga_url: str) -> 'List[Chapter]':
        soup = BeautifulSoup(self.get_str(manga_url), 'html.parser')
        ulist = soup.find('div', class_='detail_list').ul
        chapters = [Chapter('', get_text(a), 'http:'+a['href'])
                    for a in ulist.find_all('a')]
        chapters.reverse()
        return chapters

    def get_images(self, chapter_url: str) -> 'List[str]':
        soup = BeautifulSoup(self.get_str(chapter_url), 'html.parser')
        opts = soup.find('select', class_='wid60').find_all('option')
        return ['http:'+opt['value']
                for opt in opts if get_text(opt) != 'Featured']

    def get_image(self, image_url: str) -> str:
        soup = BeautifulSoup(self.get_str(image_url), 'html.parser')
        return soup.find('img', id='image')['src']


def get_downloaders() -> 'List[Downloader]':
    """Creates a list with instances of all supported downloaders.

    :return: the list of all supported downloaders.
    """
    return [NineManga('en'), NineManga('es'), NineManga('ru'),
            NineManga('de'), NineManga('it'), NineManga('br'),
            HeavenManga(), MangaReader(), MangaAll(),
            MangaDoor(), MangaNelo(), MangaHere()]



# -*- coding: utf-8 -*-
"""
.. module:: utils

This module provides utility functions and classes.

.. moduleauthor:: Asiel Díaz Benítez <asieldbenitez@gmail.com>

"""
from abc import ABC, abstractmethod
import configparser
import json
import logging
import os

from random import choice
import sys
import typing

if typing.TYPE_CHECKING:
    from typing import Callable, Iterator, List, TypeVar
    import bs4    # type: ignore
    import smd
    NoReturn = TypeVar('NoReturn')

    def _(msg: str) -> str:
        return msg


def persistent_operation(function: 'Callable') -> 'Callable':
    """Makes the decorated function to excecute again on KeyboardInterrupt
    or SystemExit exceptions.
    """
    def _wrapper(*args, **kargs):
        try:
            return function(*args, **kargs)
        except (KeyboardInterrupt, SystemExit):
            _wrapper(*args, **kargs)
            raise
    return _wrapper


class MetaFolder(ABC):

    """Abstract class representing a meta data folder."""

    data_filename = 'data.json'

    def __init__(self, path: str) -> None:
        """
        :param path: the path to the meta data folder.
        """
        self.path = path
        self._ignored = ['_ignored', 'path']

    @staticmethod
    def _load_from_folder(obj: 'MetaFolder') -> None:
        """Loads the given object is meta data"""
        with open(obj.data_file) as data_fh:
            data = json.load(data_fh)
        for key in data:
            setattr(obj, key, data[key])

    @property
    def data_file(self) -> str:
        """The path to the meta data file."""
        return os.path.join(self.path, self.data_filename)

    @staticmethod
    @abstractmethod
    def from_folder(path: str) -> 'MetaFolder':
        """Creates a ``MetaFolder`` instance from a folder path.

        :param path: path to a folder.
        :return: the created instance.
        """
        pass

    @classmethod
    def is_valid(cls, path: str) -> bool:
        """Checks whether the given path is a valid MetaFolder or not.

        :param path: path to a folder.
        :return: ``True`` if the given folder is valid, ``False``
                 otherwise.
        """
        return os.path.isfile(os.path.join(path, cls.data_filename))

    @persistent_operation
    def save_data(self) -> None:
        """Saves meta data."""
        data = {k:v for k, v in self.__dict__.items()
                if k not in self._ignored}
        with open(self.data_file, 'w') as data_fh:
            json.dump(data, data_fh)


class Chapter(MetaFolder):

    """Class representing a chapter folder."""

    data_filename = 'chapter.json'

    def __init__(self, path: str, title: str, url: str) -> None:
        """
        :param path: the path to the chapter folder.
        :param title: the chapter title.
        :param url: the chapter URL.
        """
        super().__init__(path)
        self.title = title
        self.url = url
        self.current = -1  # type: int
        self.images = []   # type: List[str]

    def __str__(self) -> str:
        return self.title

    def __repr__(self) -> str:
        return '({}, {}, {})'.format(self.title, self.url, self.current)

    @staticmethod
    def from_folder(path: str) -> 'Chapter':
        """Creates a :class:`~smd.utils.Chapter` instance from a folder path.

        :param path: the path to the chapter folder.
        :return: the created :class:`~smd.utils.Chapter` instance.
        """
        chap = Chapter(path, '', '')
        Chapter._load_from_folder(chap)
        return chap


class Config:

    """Class to manage the configuration of ``smd``"""

    def __init__(self, path: str) -> None:
        """
        :param path: the path to the configuration file.
        """
        self.path = path
        self._parser = configparser.ConfigParser()
        if self.exists():
            self.load()
        else:
            self.reset()

    def __getitem__(self, key: str):
        if key in self._parser:
            return self._parser[key]
        else:
            return self._parser['DEFAULT'][key]

    def __setitem__(self, key: str, value: str) -> None:
        self._parser['DEFAULT'][key] = value

    def exists(self) -> bool:
        """Checks if the configuration file exists.

        :return: ``True`` if the configuration file exists, ``False``
                 otherwise.
        """
        return os.path.exists(self.path)

    def load(self) -> 'List[str]':
        """Loads the configuration from disk.

        :return: a list with the loaded configuration files.
        """
        return self._parser.read(self.path)

    def reset(self) -> None:
        """Resets all configuration values to the defaults."""
        default = self._parser['DEFAULT']
        default['language'] = 'SYSTEM'
        default['manga_dir'] = '.'

    def save(self) -> None:
        """Saves the configuration to disk."""
        with open(self.path, 'w') as config_fd:
            self._parser.write(config_fd)


class ConsoleFilter(logging.Filter):

    """A filter to avoid showing exceptions stack traces text in user's
    terminal."""

    @staticmethod
    def filter(record: logging.LogRecord) -> bool:
        """Removes exception stack traces information."""
        record.exc_info = None
        record.exc_text = None  # type: ignore
        return True


class Manga(MetaFolder):

    """Class representing a manga folder."""

    data_filename = 'manga.json'

    def __init__(self, path: str, title: str, url: str, site: str) -> None:
        """
        :param path: the path to the manga folder.
        :param title: the manga title.
        :param url: the manga URL.
        :param site: the name of the downloader of this manga.
        """
        super().__init__(path)
        self.title = title
        self.url = url
        self.site = site

    def __str__(self) -> str:
        return self.title

    def __repr__(self) -> str:
        return '({}, {}, {})'.format(self.title, self.url, self.site)

    def chapters(self) -> 'Iterator[Chapter]':
        """Returns an iterator that yields chapters found in the manga
        folder"""
        for chap in os.listdir(self.path):
            path = os.path.join(self.path, chap)
            if Chapter.is_valid(path):
                yield Chapter.from_folder(path)

    @staticmethod
    def from_folder(path: str) -> 'Manga':
        """Creates a :class:`~smd.utils.Manga` instance from a folder path.

        :param path: the path to the manga folder.
        :return: the created :class:`~smd.utils.Manga` instance.
        """
        manga = Manga(path, '', '', '')
        Manga._load_from_folder(manga)
        return manga

    def get_new_chapter_path(self) -> str:
        """Generates a nonexistent chapter path.
        :return: a path to a chapter folder that don't already exists.
        """
        chap_num = 1
        path = os.path.join(self.path, str(chap_num).zfill(6))
        while os.path.exists(path):
            chap_num += 1
            path = os.path.join(self.path, str(chap_num).zfill(6))
        return path


def die(msg: str, status: int = 1) -> 'NoReturn':
    """Shows a message and terminates the program execution.

    :param msg: the message to show before terminating the application.
    :param status: the exit code.
    """
    print(msg)
    sys.exit(status)


def get_mangas(path: str) -> 'List[Manga]':
    """Gets all mangas found in the given folder.

    :param path: path to a folder.
    :return: the list of mangas found.
    """
    return [Manga.from_folder(os.path.join(path, name))
            for name in os.listdir(path)
            if Manga.is_valid(os.path.join(path, name))]


def get_text(tag: 'bs4.Tag') -> str:
    """Extracts the text from a BeautifulSoup tag.

    :param tag: the tag to extract text from.
    :return: the tag text without ``\\n`` and with trailing white spaces
             removed.
    """
    return tag.get_text().replace('\n', ' ').strip()


def mkdir(dirname: str, basename: str) -> str:
    """Tries to create a new folder ``basename`` in the folder ``dirname``
    if the name of the new folder is invalid or already exists ask the
    user to enter a new one.

    :param dirname: the parent directory of the new folder.
    :param basename: the name of the folder.
    :return: the path of the new created folder.
    """
    while True:
        path = os.path.join(dirname, basename)
        if os.path.exists(path):
            print(_("[*] ERROR - Can't create folder: '{}' already exists.")
                  .format(basename))
        else:
            try:
                os.mkdir(path)
                break
            except FileNotFoundError:
                print(_("[*] ERROR - Can't create folder '{}': Invalid name.")
                      .format(basename))
        basename = input(_("Enter a new folder name:"))
    return path


def random_ua() -> str:
    """Generates a random User-Agent HTTP header."""
    os = choice(('X11; Ubuntu; Linux x86_64;',
                  'Windows NT 10.0; Win64; x64;',
                 'Macintosh; Intel Mac OS X 10.7;'))
    browser_version = choice(('60.0', '61.0', '62.0', '59.0'))
    return 'Mozilla/5.0 ({0} rv:{1}) Gecko/20100101 Firefox/{1}' \
        .format(os, browser_version)


def select_chapters(chapters: 'List[Chapter]',
                    selectors_str: str) -> 'List[Chapter]':
    """Selects the chapters specified in the given selectors.

    :param chapters: a list of chapters.
    :param selectors: a string of comma-separated selectors.
    :return: a list of selected chapters.
    """
    if not selectors_str:
        return chapters
    selectors = selectors_str.split(',')
    chaps = set()
    ignored_chaps = set()
    for selector in selectors:
        ignore = False
        selector = selector.replace(' ', '')
        selec_str = selector[:]
        if selec_str.startswith('!'):
            ignore = True
            selec_str = selec_str[1:]
        if set(selec_str) - set('1234567890-:'):
            die(_("[*] ERROR - Invalid chapter selector: '{}'").format(selector))
        try:
            selec = selec_str.split(':')
            if selec[0]:
                i = int(selec[0])
                if i > 0:
                    selec[0] = str(i-1)
            selec_str = ':'.join(selec)
            selec_chaps = eval("chapters[{}]".format(selec_str))
            if isinstance(selec_chaps, list):
                if not selec_chaps:
                    die(_(
                        "[*] ERROR - Selector '{}' did not selected any chapters."
                    ).format(selector))
                elif ignore:
                    ignored_chaps.update(selec_chaps)
                else:
                    chaps.update(selec_chaps)
            else:
                if ignore:
                    ignored_chaps.add(selec_chaps)
                else:
                    chaps.add(selec_chaps)
        except (SyntaxError, ValueError):
            die(_("[*] ERROR - Invalid chapter selector: '{}'").format(selector))
        except IndexError:
            die(_("[*] ERROR - Chapter selector out of range: '{}'")
                .format(selector))
    if not chaps:
        chaps = set(chapters)
    return sorted(chaps - ignored_chaps, key=lambda c: chapters.index(c))


def select_mangas(mangas: 'List[Manga]', list_header: str = 'Found:',
                  prompt_msg: str = 'Select a manga',
                  multiple: bool = True) -> 'List[Manga]':
    """Lets the user select a manga from the given list, keeps asking if
    the user enters invalid option numbers.

    :param mangas: a list of mangas.
    :param list_header: the header message to show on top of the list.
    :param prompt_msg: the message prompt.
    :param multiple: if ``True`` allows to select multiple choices.
    :return: the selected manga (or mangas).
    """
    print(list_header)
    mlen = len(mangas)
    if multiple:
        mlen += 1
    dcount = len(str(mlen))
    for i, manga in enumerate(mangas, 1):
        print("{}. {}".format(str(i).rjust(dcount), manga))
    if multiple:
        print("{}. {}".format(mlen, _('[SELECT ALL]')))
    while True:
        try:
            selec_str = input(prompt_msg+" [1-{}]:".format(mlen))
            if ',' in selec_str:
                selec = [int(s) for s in selec_str.split(',')]
            else:
                selec = [int(s) for s in selec_str.split()]
            for i in selec:
                if i <= 0:
                    raise ValueError
            if multiple:
                if mlen in selec:
                    return mangas
                else:
                    return [mangas[i-1] for i in selec]
            elif len(selec) == 1 and selec[0] > 0 and selec[0] <= mlen:
                    return [mangas[selec[0]-1]]
        except ValueError:
            pass
        print(_("[*] ERROR - Invalid selection. Try again."))



# -*- coding: utf-8 -*-
"""
smd setup script.
"""

import re

from setuptools import find_packages, setup  # type: ignore

with open('README.rst', 'rt', encoding='utf8') as fh:
    README = fh.read()

with open('smd/__init__.py', 'rt', encoding='utf8') as fh:
    match = re.search(r'__version__ = \'(.*?)\'', fh.read(), re.M)
    if match:
        VERSION = match.group(1)
    else:
        raise Exception('Version not found.')

setup(
    name='smd',
    version=VERSION,
    license='GPL3+',
    author='adbenitez',
    author_email='asieldbenitez@gmail.com',
    description='Simple Manga Downloader, a tool to search and download manga for offline reading',
    long_description=README,
    long_description_content_type='text/x-rst',
    url='https://adbenitez.github.io/smd',
    classifiers=(
        'Development Status :: 4 - Beta',
        'Intended Audience :: End Users/Desktop',
        'Environment :: Console',
        'Environment :: Web Environment',
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Topic :: Games/Entertainment',
        'Topic :: Internet',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Utilities'
    ),
    keywords='download manga downloader web scraping crawler spider internet desktop app offline reading',
    project_urls={
        'Documentation': 'http://smd.readthedocs.io',
        'Donate': 'http://liberapay.com/adbenitez',
        'Say Thanks!': 'http://mastodon.social/@adbenitez',
        'Source': 'https://github.com/adbenitez/smd',
        'Tracker': 'https://github.com/adbenitez/smd/issues'
    },
    packages=find_packages(exclude=('tests*', 'docs')),
    install_requires=['beautifulsoup4'],
    python_requires='>=3.5',
    entry_points={
        'console_scripts': [
            'smd=smd:main',
        ],
    },
    include_package_data=True
)



# -*- coding: utf-8 -*-
"""
.. module:: smd

:abbr:`smd (Simple Manga Downloader)` is a tool to search and download manga
from online manga reading web sites.

.. moduleauthor:: Asiel Díaz Benítez <asieldbenitez@gmail.com>

"""
__author__ = 'adbenitez'
__version__ = '1.6.2b'

import argparse
import gettext
import os
import sys
import typing

from smd.downloader import Downloader, get_downloaders
from smd.utils import die

if typing.TYPE_CHECKING:
    from typing import Callable, List, TypeVar, Union
    import smd
    NoReturn = TypeVar('NoReturn')

    def _(msg: str) -> str:
        return msg


CONF_DIR = os.path.join(os.path.expanduser('~'), 'smd')


def keyboard_interrupt(function: 'Callable') -> 'Callable':
    """Makes the decorated function to handle keyboard interrupt gracefully."""
    def _wrapper(*args, **kargs):
        try:
            function(*args, **kargs)
        except KeyboardInterrupt:
            print(_('\n[*] Operation canceled, Sayonara! :)'))
            sys.exit()
    return _wrapper


def create_config_folder() -> None:
    """Creates the application is configuration folder, if it doesn't exists."""
    if not os.path.isdir(CONF_DIR):
        try:
            os.mkdir(CONF_DIR)
        except (FileNotFoundError, FileExistsError):
            utils.die(_(
                "[*] CRITICAL ERROR - Can't create configuration folder '{}'.")
                .format(CONF_DIR))


def download(downloaders: 'List[smd.downloader.Downloader]', manga: str,
             chapter_selectors: str, tryall: bool) -> bool:
    """Downloads the given manga using one of the given downloaders.

    :param downloaders: the list of downloaders.
    :param manga: the manga name.
    :param chapter_selectors: the comma-separated list of chapters to download.
    :param tryall: when ``True`` if a downloader fail, then tries other
                   downloaders.
    :return: ``True`` if the download succeeded, ``False`` otherwise.
    """
    downloader = downloaders.pop(0)
    if downloader.download(manga, chapter_selectors):
        return True
    else:
        downloader.logger.error(_("Download have failed :("))
        if tryall and downloaders:
            downl = select_downloader(downloaders)
            downloaders.remove(downl)
            downloaders.insert(0, downl)
            return download(downloaders, manga, chapter_selectors, tryall)
        else:
            return False


def filter_downloaders(lang: str, downloaders: 'List[smd.downloader.Downloader]') -> 'List[smd.downloader.Downloader]':
    """Selects from the given list, the downloaders with the given language.

    :param lang: the language of the downloaders to select.
    :param downloaders: a list of downloaders.
    :return: a list of downloaders with the given language.
    """
    if not lang:
        return downloaders
    lang = lang.lower()
    langs = sorted(set(d.lang for d in downloaders))
    if lang not in langs:
        print(_("[*] ERROR - Unknow language: '{}'").format(lang))
        lang = select_lang(langs)
    return [d for d in downloaders if d.lang == lang]


def get_args_parser() -> argparse.ArgumentParser:
    """Creates an arguments parser for the program is command line interface.

    :param version: the application is version number.
    :return: the arguments parser.
    """
    epilog = _("Mail bug reports and suggestions to "
               "<asieldbenitez@gmail.com>. Support this tool at "

               "http://liberapay.com/adbenitez, thanks!")
    parser = argparse.ArgumentParser(description=_("Simple Manga Downloader"),
                                     epilog=epilog)
    hmsg = _("show program's version number and exit")
    parser.add_argument('-v', '--version', help=hmsg, action='version',
                        version='%(prog)s ' + __version__)
    hmsg = _('shows program is copyright notice')
    parser.add_argument('--license', help=hmsg, action='store_true')
    hmsg = _('makes the program more verbose')
    parser.add_argument('--verbose', help=hmsg, action='store_true')
    hmsg = _("show a list of supported sites and exit")
    parser.add_argument("-l", "--list", help=hmsg, dest="list_sites",
                        action="store_true")
    hmsg = _("try to download manga from other sites if the selected site have"
             " failed, when used with option --lang, only try sites with the "
             "selected language")
    parser.add_argument("--tryall", help=hmsg, action="store_true")
    hmsg = _("use a file as input for mangas to download, the file must have a"
             " list of manga names one by line")
    parser.add_argument("-f", "--file", help=hmsg, nargs='?', const=sys.stdin,
                        type=argparse.FileType('r'))
    group = parser.add_mutually_exclusive_group()
    hmsg = _("the site from which to download")
    group.add_argument("-s", "--site", help=hmsg, metavar="SITE")
    hmsg = _("download only from sites of the selected language")
    group.add_argument("--lang", help=hmsg, metavar="LANG")
    hmsg = _("download only the given chapters, when not used all chapters are"
             " downloaded. Valid selectors are: chapter numbers (e.g. 2,15,20)"
             ", ranges (e.g. 1:10, 30:, :20), negative numbers which mean the "
             "chapter number in reverse order (e.g. -1 means the last chapter)"
             " and to ignore chapters instead of download them use '!' in "
             "front of any of the previous selectors (e.g. !5, !2:20, !-2). "
             "You can use multiple selectors separating them with a comma "
             "(e.g. '5,10,20:30,-1')")
    parser.add_argument("--chapters", help=hmsg)
    hmsg = _("the directory to use to save the mangas, by default the working "
             "directory is used")
    parser.add_argument("-d", "--directory", help=hmsg)
    hmsg = _("searches for new chapters of the given mangas")
    parser.add_argument("-u", "--update", help=hmsg, action='store_true')
    hmsg = _("continues the canceled download of the given mangas")
    parser.add_argument("-c", "--continue", help=hmsg, dest='resume',
                        action='store_true')
    hmsg = _("the name (or partial name) of the manga to search and download, "
             "or paths to manga folders when used with `-u` or `-c` options")
    parser.add_argument("mangas", help=hmsg, metavar="MANGA", nargs="*")
    return parser


def list_downloaders(downloaders: 'List[smd.downloader.Downloader]') -> None:
    """Prints the list of the given downloaders.

    :param downloaders: a list of downloaders.
    """
    print(_("Supported sites ({}):").format(len(downloaders)))
    for downloader in downloaders:
        print(" * {} ({})".format(downloader, downloader.lang))


def resume(downloaders: 'List[smd.downloader.Downloader]',
           mangas: 'Union[List[str], str]') -> None:
    """Resumes a previously canceled manga download.

    :param downloaders: the supported downloaders.
     :param mangas: a list of paths to manga folders or a path to a folder
                    where the manga folders are stored.
    """
    mangalist = []  # type: List[smd.utils.Manga]
    if isinstance(mangas, str):
        for manga in utils.get_mangas(mangas):
            for chap in manga.chapters():
                if chap.current != len(chap.images):
                    mangalist.append(manga)
                    break
        if len(mangas) == 0:
            utils.die(_("No canceled download found in '{}'.").format(mangas))
        p_msg = _('Select mangas to continue downloading')
        mangalist = utils.select_mangas(mangalist, prompt_msg=p_msg)
    else:
        mangalist = [utils.Manga.from_folder(path) for path in mangas]
    for manga in mangalist:
        for downl in downloaders:
            if downl.name == manga.site.lower():
                downl.resume(manga)
                break
        else:
            utils.die(_("[*] ERROR - Unknow site: '{}'").format(manga.site))


def select_downloader(downloaders: 'List[smd.downloader.Downloader]') -> 'smd.downloader.Downloader':
    """Lets the user choose one downloader from the give list, keeps asking if
    the user enters invalid option.

    :param downloaders: a list of downloaders.
    :return: the selected downloader.
    """
    print(_('Supported sites:'))
    dcount = len(str(len(downloaders)))
    for i, downl in enumerate(downloaders, 1):
        print("{}. {} ({})".format(str(i).rjust(dcount),
                                   downl.name, downl.lang))
    while True:
        try:
            i = int(input(_("Choose a site [1-{}]:")
                          .format(len(downloaders)))) - 1
            if 0 <= i < len(downloaders):
                break
        except ValueError:
            pass
        print(_("[*] ERROR - Invalid selection. Try again."))
    return downloaders[i]


def select_lang(langs: 'List[str]') -> str:
    """Lets the user select a language from the given list, keeps asking if
    the user enters invalid option.

    :param langs: the list of languages.
    :return: the selected language.
    """
    print(_('Available languages:'))
    dcount = len(str(len(langs)))
    for i, lang in enumerate(langs, 1):
        print("{}. {}".format(str(i).rjust(dcount), lang))
    while True:
        try:
            i = int(input(_("Choose a language [1-{}]:")
                          .format(len(langs)))) - 1
            if i >= 0 and i < len(langs):
                break
        except ValueError:
            pass
        print(_("[*] ERROR - Invalid selection. Try again."))
    return langs[i]


def set_locale(language: str, localedir: str) -> None:
    if language == 'SYSTEM':
        gettext.install('smd', localedir)
    else:
        lang = gettext.translation('smd', localedir=localedir,
                                   languages=[language], fallback=True)
        lang.install()


def set_site(site: str, downloaders: 'List[smd.downloader.Downloader]') -> None:
    """Sets the downloader of the given site as the preferred.

    :param site: the site name (same as ``smd.downloader.Downloader.name``)
    :param downloaders: the list of supported downloaders.
    """
    if site is not None:
        site = site.lower()
        for downl in downloaders:
            if downl.name == site:
                downloader = downl
                break
        else:
            print(_("[*] ERROR - Unknow site: '{}'").format(site))
            downloader = select_downloader(downloaders)
    else:
        downloader = select_downloader(downloaders)
    downloaders.remove(downloader)
    downloaders.insert(0, downloader)


def show_copyright() -> None:
    """Shows copyright notice."""
    print(_(
        "\n\tsmd --- Simple Manga Downloader.\n"
        "\tCopyright (c) 2017-2018 Asiel Díaz Benítez.\n\n"
        "\tsmd is free software: you can redistribute it and/or modify\n"
        "\tit under the terms of the GNU General Public License as published"
        " by\n"
        "\tthe Free Software Foundation, either version 3 of the License, or\n"
        "\t(at your option) any later version.\n\n"
        "\tsmd is distributed in the hope that it will be useful,\n"
        "\tbut WITHOUT ANY WARRANTY; without even the implied warranty of\n"
        "\tMERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
        "\tGNU General Public License for more details.\n\n"
        "\tYou should have received a copy of the GNU General Public License\n"
        "\tlong with smd. If not, see <http://www.gnu.org/licenses/>.\n"))


def update(downloaders: 'List[smd.downloader.Downloader]',
           mangas: 'Union[List[str], str]') -> None:
    """Updates a previously downloaded manga.

    :param downloaders: a list of supported downloaders.
    :param mangas: a list of paths to manga folders or a path to a folder
                   where the manga folders are stored.
    """
    mangalist = []  # type: List[smd.utils.Manga]
    if isinstance(mangas, str):
        mangalist = utils.get_mangas(mangas)
        if len(mangalist) == 0:
            utils.die(_("[*] ERROR - No manga found in '{}'.").format(mangas))
        mangalist = utils.select_mangas(mangalist,
                                        prompt_msg=_('Select mangas to update'))
    else:
        mangalist = [utils.Manga.from_folder(path) for path in mangas]
    for manga in mangalist:
        for downl in downloaders:
            if downl.name == manga.site.lower():
                downl.update(manga)
                break
        else:
            utils.die(_("[*] ERROR - Unknow site: '{}'").format(manga.site))


@keyboard_interrupt
def main(cmd_args: 'List[str]' = sys.argv[1:]) -> 'NoReturn':
    """Executes the command line interface of smd."""
    create_config_folder()
    config = utils.Config(os.path.join(CONF_DIR, 'smd.cfg'))
    if not config.exists():
        config.save()
    set_locale(config['language'].strip(),
               os.path.join(os.path.dirname(__file__), 'locale'))
    argparse._ = _  # type: ignore
    args = get_args_parser().parse_args(cmd_args)
    Downloader.logfile = os.path.join(CONF_DIR, __name__+'.log')
    Downloader.verbose = args.verbose
    downloaders = filter_downloaders(args.lang, get_downloaders())
    downloaders.sort(key=lambda d: d.lang)
    failed = 0
    if args.file:
        args.mangas = [m.strip() for m in args.file.readlines()
                       if m and m[0] != '#']
    if args.directory:
        os.chdir(args.directory)
    else:
        os.chdir(config['manga_dir'])
    if args.license:
        show_copyright()
    elif args.list_sites:
        list_downloaders(downloaders)
    elif args.resume:
        if args.mangas:
            resume(downloaders, args.mangas)
        else:
            resume(downloaders, args.directory)
    elif args.update:
        if args.mangas:
            update(downloaders, args.mangas)
        else:
            update(downloaders, args.directory)
    else:
        set_site(args.site, downloaders)
        if not args.mangas:
            args.mangas = [input(_("Enter manga name or text to search:"))]
        for manga in args.mangas:
            if not download(downloaders[:], manga, args.chapters,
                            args.tryall):
                failed += 1
    sys.exit(failed)
